!======================================================================== 
      subroutine denser( n     ,c_ind , r_point,  omega,        &
                         ilu   ,ilup  ,  iwork ,  space,        &
                         ierr  , mneed )

!**********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    **
!                     Indiana University                             **
!**********************************************************************

      implicit none
      include '../include/intdbl.inc'
!======================================================================== 
!
! on entry:
!========== 
!  n       = The order of the matrix A.
!  c_ind, r_point = Integer arrays. The structure of matrix A stored in 
!           CSR format.
!  levfill = Integer. Level of fill-in allowed.
!
! on return:
!=========== 
!
! ilu   = The integer pointer. iwork(ilu) is the array of ILU preconditioners.
! ilup   = Integer array of length n containing the pointers to        
!           the beginning of each row of u in the matrix ldu,ijlu. 
! ierr   = Error message with the following meaning.
!           ierr  =  0   --> successful return.
!           ierr  >  0   --> zero pivot encountered at step number ierr.
!           ierr  = -1   --> not enough storage, mneed contains the
!                        approximate additional storage needed for ldu
!
! mneed   = contains the actual number of elements in ldu, or the amount
!       of additional storage needed for ldu
!
! work arrays:
!=============
! iwork      = whole working space. All the working arrays are subarray
!          of this array.
! rowll      = index to the integer array of length n containing pointers 
!          to implement a linked list for the fill-in elements.
!
! calling routines:
!==================
!
! external functions:
!====================
!
! note:
!======
! the columns of matrix A shall be sorted in increase order.
!======================================================================== 
!
!----------------------------------------------------------------------- 
      integer n,c_ind(*), r_point(*)
      integer ilu,ilup,levfill
      integer lastcol, rowll, levels, remain, nzlu
      integer iwork(*), space
      integer ierr,   mneed
      integer i, nnz_ilu
      double precision ::  omega
!
!
! Compute pointers into work array.  This version uses two arrays
! dependent on the (unknown in advance) size nzlu, ilu(1:nzlu) and
! levels(1:nzlu).  To handle this, first the data space for the
! work arrays in symbfac is allocated, then half the remainder is
! assigned to ilu, the rest to levels.  The partitioning of the 
! integer work array is as:
!
!         [ilup   ilu      ...      levels    rowll lastcol],
!         [-n-----nzlu-----...-------nzlu-------n------n---]
!
!  The index of the last entry in the work array is space, and
!  rowll and lastcol start at the far end of the work array.
!  Note that these are not necessarily aligned on double word 
!  boundaries.
!
      lastcol = space - n + 1
      rowll = lastcol - n
!
! Set ilup to the far left of the avaiable work array space.
!
      ilup = 1
      ilu  = ilup + n + mod(n,intdbl)
!
! Of space remaining, can allocate half to ilu and half to levels.
! Compute the halfway dividing mark "remain".  Note that
!    growth available = total - lastcol  - rowll  - ilup 
!
      remain = (space - 3*n) / 2
!
! Now levels = storage for ilup + half remaining words.
!
      levels = n                + remain
!
! If remain is nonpositive, estimate 4n as needed storage, 2n for lu
! and 2n for ilu: this will handle a diagonal preconditioner.
!
      if (remain .le. 0) then
         ierr = -1
         mneed = 4*n + 1
         return
      end if
!
! Perform symbolic factorization:
!
      levfill = int(omega)
      call symbfac(n     , c_ind ,r_point,levfill,remain ,nzlu   ,     &
           iwork(ilu)   ,iwork(ilup), iwork(lastcol),iwork(levels),    &
           iwork(rowll), ierr  ,mneed)

        if (ierr .ne. 0) then
           return
        endif
!
!  Merge the pattern with the pattern of A and set index of the pattern
!  to negative. The elements in A only (not in the pattern) are kept positive.
!  Because the pattern is denser than A, only need to set index to be negative.
!  Note that nnz_ilu shown here is actually the number of nonzeros from the
!  off-diagonal.
!
        nnz_ilu = iwork(ilu+n)-n-2
        do i = ilu+n+1, ilu+n+nnz_ilu
            iwork(i) = -iwork(i)
        enddo
1000    continue
        return
        end
