!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
      subroutine ecimgs(n,a,c_ind,r_point,relax,ilu,alu,ilup,iwork, &
                      maxspace, ierr, mneed)

!*********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    *
!                     Indiana University                             *
!*********************************************************************

      implicit none
      include '../include/intdbl.inc'
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
!                                                                       
!    Extended CIMGS for general sparse matrix.                      
!                                       
!   This routine performs ECIMGS with different pattern of target   
!   sparsity patterns. It allows the sparsity pattern be sparser    
!   than the sparsity pattern of the given matrix A or be denser    
!   than the sparsity pattenr of A.                 
!                                   
!   ECIMGS requires more storage than the space for storing the 
!   non zeros of  the factors. Amount of intermediate storage   
!   is needed.                              
!                                   
!   For More information of ECIMGS, see                     
!                                   
!     Xiaoge Wang                                           
!     Department of Computer Science                        
!     Indiana University                            
!     email: xwang@cs.indiana.edu                       
!     Fri Aug  5 15:20:35 EST 1994                                      
!                                                                       
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
!
! parameters
!-----------
!
! on entry:
!========== 
! n       : integer. the dimension of the matrix A.
!
! a,c_ind,r_point : matrix A stored in compressed sparse row format.
!
! relax   : pattern selection relaxation parameter.
!       relax = 0.0     use non zero pattern of matrix A
!       0.0 < relax < 1.0 use sparser pattern. 
!       relax = 1.0     use diagonal pattern
!       relax >1.0      use denser pattern (int(relax) level of fillin)
!
! on return:
!=========== 
!
! ilu, ilup, alu : pointers to the position of preconditioners L and U
!    in MSR format in the workarray iwork.
!                                                                        
! ierr is an error flag:
!        ierr  = -i --> near zero pivot in step i.
!        ierr  = 0  --> no error is detected.
!        ierr  = 1  --> not enough storage; check mneed.
!
! mneed   = the number of elements in iwork that have been used
!           for the factorization if it is successful, or the number
!       of additional storage locations needed for lu otherwise.
!
! Work arrays:
!=============
!
! iwork     = integer work array of the whole available space.
!
! maxspace      = size of iwork
!
! Storage requirements:
!======================
!
!  Warnings:
!===========
!
! 1) A must have nonzero diagonal entries, at least.  This does not assure
!    completion of the factorization, however.
!
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
      integer  n, r_point(*), c_ind(*)
      double precision ::   a(*)
      double precision ::   relax
      integer  iwork(*), maxspace
      integer  ierr, used, mneed
      integer  ilu, alu, ilup
      integer  nzlu, lstrow, alu_new
!
! Initialize the parameters
!
      ierr  = 0
      mneed = 0
      used  = 0
!
! If parameter of fills is negative, set to 0
! 
      relax = max( relax, 0.0d0)
!
! Perform symbolic factorization:
! on return, iwork will be taken by ilup, ilu. mneed gives the
! size of used space. Total available space after symbolic analysis
! is then (maxspace - mneed).
!
      call symbgs(n, a, c_ind , r_point , relax,                &
                   ilu   , ilup  ,  iwork(1)  , maxspace,       &
                   ierr  , mneed)
      if ( ierr .ne. 0 ) then
!         print*, 'symbfac fail due to the space shortage', maxspace
      ierr = 1
          return
      endif
!          
! Perform numerical factorization: more space are needed for numerical
! factorization. n integer for working space and nzlu double precision :: for the
! values of the elements in LU. 
!
      nzlu = iwork(ilu+n)-1
      used = ilu + nzlu - 1
      if (maxspace-used .lt. intdbl*nzlu+n) then
         ierr =  1
         mneed = intdbl*nzlu+n - (maxspace-used)
         return
      endif
      alu =  used + 1
      alu =  alu + mod(alu-1,intdbl)
      lstrow = alu + intdbl*nzlu
!      mneed = max(lstrow + n - 1, mneed)
      mneed = n + (1 + intdbl)*nzlu
!
! use numfac without extra real type working space
!
      call  numgs(n, r_point, c_ind, a, iwork(ilu),iwork(ilup),     &
              iwork(alu) , iwork(lstrow), ierr) 
      nzlu = iwork(ilu+n)-1
      if (ierr .ne. 0) then
      print*, 'error in numerical phase of ECIMGS'
      ierr = -ierr
      return
      endif
!
! Collect space
!
      alu_new = ilu+nzlu
      alu_new = alu_new + mod(alu_new-1,intdbl)
      call collect( iwork(alu), iwork(alu_new), nzlu, 1)
      alu = alu_new
!
      return
      end
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
    subroutine collect(a1, a2, n, flag)
!
!   mv real array a1 to array a2. the length of the arrays
!   is n. flag indicate the direction of moving. when flag=0
!   move right. when flag=1, move to left
!   this subroutine is usd to collect the space in a global array.
!
    implicit none
    integer n, flag, i
    double precision :: a1(n), a2(n)
    
    if (flag.eq.0) then
!   move right: start from tail of the array
        do i = n, 1, -1
        a2(i) = a1(i)
        enddo
    elseif(flag.eq.1) then
!   move left: start form head of the array
        do i = 1, n
        a2(i) = a1(i)
        enddo
    else
        print*, 'direction flag error in subroutin COLLECT'
    endif
    return
    end
