!=======================================================================*
      subroutine cscgstab(A,colind,rwptr,n,                     &
                          x,r,b,                                &
                          tolrnc,reltv,maxits,ierr,force,       &
                          work,wksz,                            &
                          lu,jlu,uptr,pcmeth, iunit,instlvl)

!-----------------------------------------------------------------------*
!                       *** CSCGSTAB ***                
!  The composite step stabilized biconjugate gradient algorithm of Chan
!  and Szego.  This is from the article in the November 1996 issue of 
!  SIAM J. Sci. Comp., and may not be completely correct.  Furthermore,
!  the preconditioning it performs is a very simple-minded left pc 
!  without any sophisticated norm/error checking.
!-----------------------------------------------------------------------*
!
!  This routine is just a wrapper for the subroutine cscgs.f, which is
!  in the same file below.  See it for a description of all the variables.
!  This wrapper carves up the workspace vector work() and then calls
!  cscgs, so the only error code relevant is ierr < 0, indicating not
!  enough memory.
!
!======================================================================
!  Routines
!  --------
!
!  cscgs   : This is the actual work routine.  See it for explanation
!            of the variables used.
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Sun Nov 10 13:13:45 EST 1996
!
!======================================================================

!     -------------------------
!     Declaration of arguments:
!     -------------------------
      implicit none

      integer n
      double precision :: A(*)
      double precision :: x(*), b(*), r(*)
      double precision :: work(*)
      double precision :: tolrnc(*)
      double precision :: lu(*)

      integer colind(*), rwptr(*)
      integer wksz
      integer maxits, ierr, need
      integer jlu(*), uptr(*)
      integer pcmeth, iunit(*), instlvl
      logical reltv, force

!     ------------------------------
!     Pointers to work array vectors
!     ------------------------------
      integer rtilde, p, e, q, c, u, y, d, rhat, ehat
      integer s, t, w, v, rhat2, ehat2, zt, wt, wins1

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ---------------------------
!     Check the available storage
!     ---------------------------

      need = 18
      if (instlvl .ge. 0) need = need + 2

      if (n*need .gt. wksz) then
         ierr = wksz - n*need
         maxits = 0
         return
      end if

!     -----------------------------------------
!     Set up pointers for vectors in work array
!     -----------------------------------------
      rtilde = 1
      p = rtilde + n
      e = p + n
      q = e + n
      c = q + n
      u = c + n
      y = u + n
      d = y + n
      rhat = d + n
      ehat = rhat + n
      s = ehat + n
      t = s + n
      w = t + n
      v = w + n
      rhat2 = v + n
      ehat2 = rhat2 + n
      zt = ehat2 + n
      wt = zt + n
      wins1 = wt + n

      call cscgs(A,colind,rwptr,n,                                      &
                x,r,b,                                                  &
                work(rtilde),work(p),work(e),work(q),work(c),           &
                work(u),work(y),work(d),work(rhat),work(ehat),work(s),  &
                work(t),work(w),work(v),work(rhat2),work(ehat2),        &
                work(zt),work(wt),work(wins1),                          &
                tolrnc,reltv,maxits,ierr,force,                         &
                lu,jlu,uptr,pcmeth, iunit,instlvl)

      return

!================= End of CSCGSTAB =====================================

      end


!=======================================================================*
      subroutine  cscgs(A,colind,rwptr,n,                               &
                        x,r,b,                                          &
                        rtilde,p,e,q,c,u,y,d,rhat,ehat,s,t,w,v,         &
                        rhat2,ehat2,                                    &
                        zt, wt, wins1,                                  &
                        tolrnc,reltv,maxits,ierr,force,                 &
                        lu,jlu,uptr,pcmeth, iunit,instlvl)

!-----------------------------------------------------------------------*
!                         *** CSCGSTAB ***
!  The composite step stabilized biconjugate gradient algorithm.
!
!-----------------------------------------------------------------------*
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  rtilde,p,e,q,c,u,y,d,rhat,ehat,s,t,w,v,rhat2,ehat2 :
!                                      Cast of thousands of vectors.
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL.  Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tol is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tol is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!  force   : Logical parameter to force the method to always use one-step.
!                 force = .true.  ---> use one step always
!                 force = .false. ---> use composite step 
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by cscgstab, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  urnorm  : Unpreconditioned initial residual.  Used for tests to see
!            if initial residual is good enough, and if preconditioner
!            is no good (see pdr above).
!
!  wins1   : vector of length 2n used by instrumentation routines.
!            Can be null if instlvl < 0.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!           ierr = 5  --> omega1 small divisor 
!           ierr = 6  --> mu small divisor
!           ierr = 7  --> omegah small divisor
!           ierr = 8  --> omega2 small divisor
!           ierr = 9  --> sigma small divisor
!           ierr = 10 --> mu (second location) small divisor
!           ierr = 11 --> rho small divisor
!           ierr = 12 --> delta small divisor
!           ierr = 13 --> gamma small divisor
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, daxpy, dcopy, dnrm2
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Sun Nov 10 09:34:41 EST 1996
!
!======================================================================

!     -------------------------
!     Declaration of arguments:
!     -------------------------
      implicit none

      integer n
      double precision :: A(*)
      double precision :: x(n), b(n), r(n)
      double precision :: rtilde(n),p(n),e(n),q(n),c(n),u(n),y(n)
      double precision :: d(n),rhat(n),ehat(n),s(n),t(n),w(n),v(n),z
      double precision :: rhat2(n),ehat2(n), zt(n), wt(n)
      double precision :: wins1(*)
      double precision :: tolrnc(*)
      double precision :: lu(*)

      integer colind(*), rwptr(*)
      integer maxits, ierr
      integer jlu(*), uptr(*)
      integer pcmeth, iunit(*), instlvl
      logical reltv, force

!     ------------------
!     External functions
!     ------------------
      double precision :: ddot, dnrm2
      external ddot, dnrm2

!     -------------
!     Local storage 
!     -------------
      logical first
      logical onestep
      integer i, its, resets, k
      integer steps(2)
      double precision :: res, tol
      double precision :: rho, phi, mu, sigma, omega1, omega2, psi, omegah
      double precision :: a11, a12, a21, a22, b1, b2, alpha1, alpha2
      double precision :: delta, gamma1, gamma2, nu2, beta1, beta2, nutilde
      double precision :: SMALL, urnorm, pdr
      double precision :: temp1, temp2

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      pdr = 1.0d0/epsilon(SMALL)

!----------------------------------------------------------------------

!     ---------------------
!     Initialize iterations
!     ---------------------
      steps(1) = 0
      steps(2) = 0
      resets = 0
      first = .true.
      tol = tolrnc(1)
      its = 0
      ierr = 0

!     ------------------
!     r = A*x; r = b - r
!     ------------------
      call bmux(n, n,x,p,A,colind,rwptr)
      do i = 1,n
         p(i) = b(i) - p(i)
      end do
      urnorm = ddot(n,p,1,p,1)
      call lusl(n, p, r, lu, jlu, uptr, pcmeth, A,colind,rwptr)
      do i = 1,n
         p(i) = r(i)
         rtilde(i) = r(i)
      end do

!     -------------------------------------------------------------
!     Check to see if preconditioned residual is too large relative
!     to unpreconditioned residual:
!     -------------------------------------------------------------
      if (ddot(n,r,1,r,1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

      rho = ddot(n, rtilde, 1, r, 1)
      phi = dnrm2(n, r, 1)

!     --------------
!     e = A*r; q = e
!     --------------
      call bmux(n, n,r,e,A,colind,rwptr)
      call lusl(n, e, q, lu, jlu, uptr, pcmeth, A,colind,rwptr)
      call dcopy(n,q,1,e,1)
      mu = 1

!     -------------------------------------------------
!     Adjust tolerance if relative residual to be used:
!     -------------------------------------------------
      if (reltv) then
          tol = tol*phi
          tolrnc(1) = tolrnc(1)*phi
      end if

      if (phi .lt. tol) go to 900


!=======================================================================
!     Begin main loop of iterations.  Loop is equivalent to
!      do while (phi .gt. tol) .and. (its .le. maxits)
!=======================================================================

 2       continue
         call instr(A,colind,rwptr,n, x,b, wins1, lu,jlu,uptr,      &
                     pcmeth,first,its, iunit,instlvl)

         a11 = ddot(n,rtilde,1,q,1) 
         sigma = a11*mu

!        ------
!        c = Aq
!        ------
         call bmux(n, n,q,zt,A,colind,rwptr)
         call lusl(n, zt, c, lu, jlu, uptr, pcmeth, A,colind,rwptr)

         do k = 1, n
            u(k) = sigma*r(k) - rho*q(k)
            y(k) = sigma*e(k) - rho*c(k)
         end do 

!        ------
!        d = Ay
!        ------
         call bmux(n, n,y,zt,A,colind,rwptr)
         call lusl(n, zt, d, lu, jlu, uptr, pcmeth, A,colind,rwptr)

!        ----------------
!        omega1 = y'u/y'y
!        ----------------
         temp1 = 0.0d0
         temp2 = 0.0d0
         do k = 1, n
            temp1 = temp1 + y(k)*u(k)
            temp2 = temp2 + y(k)*y(k)
         end do 
         if (temp2  .le. SMALL) go to 930
         omega1 = temp1/temp2

!        ----------------------------------------------------------
!        rhat = u - omega1*y; ehat = y - omega1*d; psi = || rhat ||
!        ----------------------------------------------------------
         psi = 0.0d0
         do k = 1, n
            rhat(k) = u(k) - omega1*y(k)
            ehat(k) = y(k) - omega1*d(k)
            psi = psi + rhat(k)*rhat(k)
         end do 
         psi = sqrt(psi)

!        ---------------------------------------------
!        Shortcircuit the two step stuff if specified.
!        ---------------------------------------------
         if (force) then
            onestep = .true.
            go to 40
         end if 

         onestep = ( psi .lt. abs(sigma)*phi )
         if (.not. onestep) then
            a12 = 0.0d0
            a22 = 0.0d0
            a21 = 0.0d0
            b2  = 0.0d0
            do k = 1, n
               a12 = a12 + rtilde(k)*y(k)
               a21 = a21 + rtilde(k)*c(k)
               a22 = a22 + rtilde(k)*d(k)
               b2  = b2  + rtilde(k)*e(k)
            end do 
            delta = a11*a22 - a12*a21
            if (abs(mu)  .le. SMALL) go to 940 
            b1 = rho/mu
            alpha1 = a22*b1 - a12*b2
            alpha2 = -a21*b1 + a11*b2
            do k = 1, n
              s(k) = delta*r(k) - alpha1*q(k) - alpha2*y(k)
              t(k) = delta*e(k) - alpha1*c(k) - alpha2*d(k)
            end do 

            temp1 = 0.0d0
            temp2 = 0.0d0
            do k = 1, n
               temp1 = temp1 + t(k)*s(k)
               temp2 = temp2 + t(k)*t(k)
            end do 
            if (abs(temp2)  .le. SMALL) go to 950
            omegah = temp1/temp2

            nutilde = 0.0d0
            do k = 1, n
               wt(k)  = s(k) - omega1*t(k)
               nutilde = nutilde + (s(k) - omegah*t(k))**2
            end do 

            nutilde = sqrt(nutilde) 
            if (abs(delta)*psi  .lt.  abs(sigma)*nutilde) then
               onestep = .true.
            else
               call bmux(n, n,t,zt,A,colind,rwptr)
               call lusl(n,zt,v,lu,jlu,uptr,pcmeth,A,colind,rwptr)
               call bmux(n, n,v,zt,A,colind,rwptr)
               call lusl(n,zt,w,lu,jlu,uptr,pcmeth,A,colind,rwptr)
               temp1 = 0.0d0
               temp2 = 0.0d0
               do k = 1, n
                  z = t(k) - omega1*v(k)
!                  temp1 = temp1 + z*u(k)
                  temp1 = temp1 + z*wt(k)
                  temp2 = temp2 + z*z
               end do 
               if (abs(temp2)  .le. SMALL) go to 960
               omega2 = temp1/temp2

               gamma1 = -(omega1 + omega2)
               gamma2 = omega1 * omega2
               nu2 = 0.0d0
               do k = 1, n
                  rhat2(k) = s(k) + gamma1*t(k) + gamma2*v(k)
                  ehat2(k) = t(k) + gamma1*v(k) + gamma2*w(k)
                  nu2 = nu2 + rhat2(k)*rhat2(k)
               end do 
               nu2 = sqrt(nu2)
               onestep = ( abs(delta)*psi .lt. abs(sigma)*nu2)
            end if 
         end if 

 40      continue

         if (onestep) then
!        =======================
!        Perform one-step CGSTAB
!        =======================
            if (abs(sigma)  .le. SMALL) go to 970
            steps(1) = steps(1) + 1
            temp1 = 1.0d0/sigma
            temp2 = 0.0d0
            do k = 1, n
               r(k) = temp1 * rhat(k)
               e(k) = temp1 * ehat(k)
               x(k) = x(k) + temp1 * (rho*p(k) + omega1*u(k))
               temp2 = temp2 + r(k)*rtilde(k)
            end do
            phi = psi*temp1
            if (abs(omega1)  .le. SMALL) go to 980
            mu = mu*rho*temp1/omega1
            temp2 = mu*temp2

            if (abs(rho)  .le. SMALL) go to 990
            beta1 = temp2/rho
            rho = temp2
            do k = 1, n
               p(k) = r(k) + beta1*(p(k) - omega1*q(k))
               q(k) = e(k) + beta1*(q(k) - omega1*c(k))
            end do 
            its = its + 1
         else
!        =======================
!        Perform two-step CGSTAB
!        =======================
            if (abs(delta)  .le. SMALL) go to 995
            steps(2) = steps(2) + 1
!            Write(*,*) 'Two step iteration: ', its
            temp1 = 1.0d0/delta
            temp2 = 0.0d0
            b1 = 0.0d0
            b2 = 0.0d0
            phi = nu2*temp1
            do k = 1, n
               r(k) = temp1 * rhat2(k)
               e(k) = temp1 * ehat2(k)
               x(k) = x(k) + temp1 *    &
                (alpha1*p(k) + alpha2*u(k) - gamma1*s(k) - gamma2*t(k))
               temp2 = temp2 + rtilde(k)*r(k)
               b1 = b1 + rtilde(k)*t(k)
               b2 = b2 + rtilde(k)*v(k)
            end do 
            if (abs(gamma2)  .le. SMALL) go to 997
            mu = -mu*temp1*alpha2*rho/gamma2
            rho = temp2*mu
            beta1 = temp1*(a22*b1 - a12*b2)
            beta2 = temp1*(-a21*b1 + a11*b2)
            do k = 1, n
               p(k) = r(k) - beta1*(p(k) + gamma1*q(k) + gamma2*c(k))   &
                           - beta2*(u(k) + gamma1*y(k) + gamma2*d(k))
            end do 
            call bmux(n, n,p,zt,A,colind,rwptr)
            call lusl(n,zt,q,lu,jlu,uptr,pcmeth,A,colind,rwptr)
            its = its + 2
         end if

         if (its .ge. maxits) go to 920

         if (dnrm2(n,r,1) .lt. tol) go to 910

        go to 2

!=======================================================================
!           Stopping conditions
!=======================================================================


!     ===================== 
!     (Almost) exact guess 
!     ===================== 
 900  continue
      ierr = 1
      maxits = 0
      return

!     ========================================= 
!     Possible convergence; check true residual 
!     ========================================= 
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 999
      end if
      call bmux(n,n,x,rhat,A,colind,rwptr)
      do i = 1, n
         rhat(i) = b(i) - rhat(i)
      end do
      res = dnrm2(n,rhat,1)
      if (res .gt. tolrnc(1)) then
          tol = tol*tolrnc(1)/res
!          write(6,*)'>>>>> tolerance reset to ', tol,'  <<<<<<'
          go to 2
      end if
      go to 999

!     ======================================== 
!     No convergence in max allowed iterations 
!     ======================================== 
 920  continue
      ierr = 3
      go to 999

!     ================================ 
!     omega1 computation small divisor 
!     ================================ 
 930  continue
      ierr = 5
      go to 999

!     ====== 
!     mu = 0 
!     ====== 
 940  continue
      ierr = 6
      go to 999

!     ================================ 
!     omegah computation small divisor
!     ================================ 
 950  continue
      ierr = 7
      go to 999

!     ================================ 
!     omega2 computation small divisor
!     ================================ 
 960  continue
      ierr = 8
      go to 999

!     =============================== 
!     sigma computation small divisor
!     =============================== 
 970  continue
      ierr = 9
      go to 999

!     =================================== 
!     second mu computation small divisor
!     =================================== 
 980  continue
      ierr = 10
      go to 999

!     ============================= 
!     rho computation small divisor
!     ============================= 
 990  continue
      ierr = 11
      go to 999

!     =============================== 
!     delta computation small divisor
!     =============================== 
 995  continue
      ierr = 12
      go to 999

!     =============================== 
!     gamma computation small divisor
!     =============================== 
 997  continue
      ierr = 13
      go to 999

!     =================================== 
!     Final instrumentation call and exit
!     =================================== 
 999  continue
      maxits = its
      Write(6,*) 'One step iterations: ', steps(1)
      Write(6,*) 'Two step iterations: ', steps(2)
      call instr(A,colind,rwptr,n, x,b, wins1, lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!================= End of CSCGS ========================================

      end
